@@ -131,8 +131,6 @@ struct _GrdRdpPipeWireStream
   uint32_t src_node_id;
 
   struct spa_video_info_raw spa_format;
-
-  gboolean destroyed;
 };
 
 G_DEFINE_TYPE (GrdRdpPipeWireStream, grd_rdp_pipewire_stream,
@@ -838,9 +836,6 @@ on_stream_process (void *user_data)
   struct pw_buffer *last_frame_buffer = NULL;
   struct pw_buffer *next_buffer;
 
-  if (stream->destroyed)
-    return;
-
   while ((next_buffer = pw_stream_dequeue_buffer (stream->pipewire_stream)))
     {
       struct spa_meta_header *spa_meta_header;
@@ -1158,21 +1153,15 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
   GrdContext *context = grd_session_get_context (session);
   GrdEglThread *egl_thread;
 
-  stream->destroyed = TRUE;
-
+  /* Setting a PipeWire stream inactive will wait for the data thread to end */
   if (stream->pipewire_stream)
-    pw_stream_flush (stream->pipewire_stream, false);
+    pw_stream_set_active (stream->pipewire_stream, false);
 
   egl_thread = grd_context_get_egl_thread (context);
   if (egl_thread && stream->pipewire_stream)
     sync_egl_thread (egl_thread);
 
-  /*
-   * We can't clear stream->pipewire_stream before destroying it, as the data
-   * thread in PipeWire might access the variable during destruction.
-   */
-  if (stream->pipewire_stream)
-    pw_stream_destroy (stream->pipewire_stream);
+  g_clear_pointer (&stream->pipewire_stream, pw_stream_destroy);
 
   g_clear_pointer (&stream->pipewire_core, pw_core_disconnect);
   g_clear_pointer (&stream->pipewire_context, pw_context_destroy);
